"""Module to check flatness of ligand substructures."""

from __future__ import annotations

from collections.abc import Iterable
from copy import deepcopy
from typing import Any

import numpy as np
from numpy import ndarray as Array
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolfiles import MolFromSmarts
from rdkit.Chem.rdmolops import SanitizeMol

flat_rings = {
    "aromatic_5_membered_rings_sp2": "[ar5^2]1[ar5^2][ar5^2][ar5^2][ar5^2]1",
    "aromatic_6_membered_rings_sp2": "[ar6^2]1[ar6^2][ar6^2][ar6^2][ar6^2][ar6^2]1",
}
flat_bonds = {"trigonal_planar_double_bonds": "[C;X3;^2](*)(*)=[C;X3;^2](*)(*)"}
flat = flat_rings | flat_bonds
nonflat = {
    # "non-aromatic_5_membered_rings": "[C,O,S,N;R1]~1[C,O,S,N;R1][C,O,S,N;R1][C,O,S,N;R1][C,O,S,N;R1]1",
    # the above matches flat 5-membered rings in O2U and AWJ
    "non-aromatic_6_membered_rings": "[C,O,S,N;R1]~1[C,O,S,N;R1][C,O,S,N;R1][C,O,S,N;R1][C,O,S,N;R1][C,O,S,N;R1]1",
    "non-aromatic_6_membered_rings_db03_0": "[C;R1]~1[C;R1][C,O,S,N;R1]~[C,O,S,N;R1][C;R1][C;R1]1",
    "non-aromatic_6_membered_rings_db03_1": "[C;R1]~1[C;R1][C;R1]~[C;R1][C,O,S,N;R1][C;R1]1",
    "non-aromatic_6_membered_rings_db02_0": "[C;R1]~1[C;R1][C;R1][C,O,S,N;R1]~[C,O,S,N;R1][C;R1]1",
    "non-aromatic_6_membered_rings_db02_1": "[C;R1]~1[C;R1][C,O,S,N;R1][C;R1]~[C;R1][C;R1]1",
}
_empty_results = {
    "results": {
        "num_systems_checked": np.nan,
        "num_systems_passed": np.nan,
        "max_distance": np.nan,
        "flatness_passes": np.nan,
    }
}


def check_flatness(
    mol_pred: Mol,
    threshold_flatness: float = 0.1,
    flat_systems: dict[str, str] = flat,
    check_nonflat: bool = False,
) -> dict[str, Any]:
    """Check whether substructures of molecule are flat.

    Args:
        mol_pred: Molecule with exactly one conformer.
        threshold_flatness: Maximum distance from shared plane used as cutoff. Defaults to 0.1.
        flat_systems: Patterns of flat (or non-flat) systems provided as SMARTS. Defaults to
             5 and 6 membered aromatic rings and carbon sigma bonds.
        check_nonflat: Whether to check the ring non-flatness instead of flatness.
            Turns (flatness <= threshold_flatness) to (flatness >= threshold_flatness).

    Returns:
        PoseBusters results dictionary.
    """
    mol = deepcopy(mol_pred)
    # if mol cannot be sanitized, then rdkit may not find substructures
    try:
        assert mol_pred.GetNumConformers() > 0, "Molecule does not have a conformer."
        SanitizeMol(mol)
    except Exception:
        return _empty_results

    planar_groups = []
    types = []
    for flat_system, smarts in flat_systems.items():
        match = MolFromSmarts(smarts)
        atom_groups = list(mol.GetSubstructMatches(match))
        planar_groups += atom_groups
        types += [flat_system] * len(atom_groups)

    # calculate distances to plane and check threshold
    coords = [_get_coords(mol, group) for group in planar_groups]
    max_distances = [float(_get_distances_to_plane(X).max()) for X in coords]
    if not check_nonflat:
        flatness_passes = [bool(d <= threshold_flatness) for d in max_distances]
        extreme_distance = max(max_distances) if max_distances else np.nan
    else:
        flatness_passes = [bool(d >= threshold_flatness) for d in max_distances]
        extreme_distance = min(max_distances) if max_distances else np.nan
    details = {
        "type": types,
        "planar_group": planar_groups,
        "max_distance": max_distances,
        "flatness_passes": flatness_passes,
    }

    results = {
        "num_systems_checked": len(planar_groups),
        "num_systems_passed": sum(flatness_passes),
        "max_distance": extreme_distance,
        "flatness_passes": all(flatness_passes) if len(flatness_passes) > 0 else True,
    }

    return {"results": results, "details": details}


def _get_distances_to_plane(X: Array) -> Array:
    """Get distances of points X to their common plane."""
    # center points X in R^(n x 3)
    X = X - X.mean(axis=0)
    # singular value decomposition
    _, _, V = np.linalg.svd(X)
    # last vector in V is normal vector to plane
    n = V[-1]
    # distances to plane are projections onto normal
    d = np.dot(X, n)
    return d


def _get_coords(mol: Mol, indices: Iterable[int]) -> Array:
    return np.array([mol.GetConformer().GetAtomPosition(i) for i in indices])



from __future__ import annotations

from rdkit.Chem.rdmolfiles import MolFromPDBFile

from posebusters.modules.distance_geometry import _two_bonds_to_angle, check_geometry


def test_check_geometry(mol_pm2):
    out = check_geometry(mol_pm2)
    assert out["results"]["bond_lengths_within_bounds"] is True
    assert out["results"]["bond_angles_within_bounds"] is True
    assert out["results"]["no_internal_clash"] is True


def test_two_bonds_to_angle():
    assert _two_bonds_to_angle((0, 1), (0, 1)) is None
    assert _two_bonds_to_angle((0, 1), (1, 0)) is None
    assert _two_bonds_to_angle((0, 1), (1, 2)) == (0, 1, 2)
    assert _two_bonds_to_angle((0, 1), (2, 1)) == (0, 1, 2)
    assert _two_bonds_to_angle((1, 0), (1, 2)) == (0, 1, 2)
    assert _two_bonds_to_angle((1, 0), (2, 1)) == (0, 1, 2)
    assert _two_bonds_to_angle((0, 1), (2, 3)) is None


def test_molecule_issue_80():
    """Test sanitizable molecule from issue 80 that causes a RunTime error in GetMoleculeBoundsMatrix."""

    mol = MolFromPDBFile("tests/conftest/mol_issue_80.pdb", removeHs=False, sanitize=True)
    assert mol is not None
    # should not raise an error
    check_geometry(mol)



"""Command line interface for PoseBusters."""

from __future__ import annotations

import argparse
import logging
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Any, TextIO

import pandas as pd
from rdkit.Chem.rdchem import Mol
from yaml import safe_load

from . import __version__
from .posebusters import PoseBusters
from .tools.formatting import create_long_output, create_short_output

logger = logging.getLogger(__name__)


def main() -> None:
    """Safe entry point for PoseBusters from the command line."""
    parser = _parse_args(sys.argv[1:])
    try:
        bust(**vars(parser))
    except Exception as e:
        logger.error(e)


def bust(  # noqa: PLR0913
    mol_pred: list[Path | Mol] = [],
    mol_true: Path | Mol | None = None,
    mol_cond: Path | Mol | None = None,
    table: Path | None = None,
    outfmt: str = "short",
    output: Path | TextIO = sys.stdout,
    config: Path | None = None,
    no_header: bool = False,
    full_report: bool = False,
    top_n: int | None = None,
    max_workers: bool = False,
    chunk_size: int | None = None,
) -> None:
    """PoseBusters: Plausibility checks for generated molecule poses."""
    if table is None and len(mol_pred) == 0:
        raise ValueError("Provide either MOLS_PRED or TABLE.")

    if table is not None:
        # run on table
        file_paths = pd.read_csv(table, index_col=None)
        mode = _select_mode(config, file_paths.columns.tolist())
        posebusters = PoseBusters(mode, top_n=top_n, max_workers=max_workers, chunk_size=chunk_size)
        posebusters.file_paths = file_paths
        posebusters_results = posebusters._run()
    else:
        # run on single input
        d = {k for k, v in dict(mol_pred=mol_pred, mol_true=mol_true, mol_cond=mol_cond).items() if v}
        mode = _select_mode(config, d)
        posebusters = PoseBusters(mode, top_n=top_n, max_workers=max_workers, chunk_size=chunk_size)
        cols = ["mol_pred", "mol_true", "mol_cond"]
        posebusters.file_paths = pd.DataFrame([[mol_pred, mol_true, mol_cond] for mol_pred in mol_pred], columns=cols)
        posebusters_results = posebusters._run()

    if isinstance(output, Path):
        output = open(Path(output), "w", encoding="utf-8")

    for i, (k, v) in enumerate(posebusters_results):
        results = posebusters._make_table({k: v}, posebusters.config, full_report=full_report)
        output.write(_format_results(results, outfmt, no_header, i))


def _parse_args(args: list[str]) -> argparse.Namespace:
    desc = "PoseBusters: Plausibility checks for generated molecule poses."
    parser = argparse.ArgumentParser(description=desc, add_help=False)

    # Create two argument groups
    in_group = parser.add_argument_group(title="Input")
    out_group = parser.add_argument_group(title="Output")
    cfg_group = parser.add_argument_group(title="Configuration")
    inf_group = parser.add_argument_group(title="Information")

    # input
    help = "molecule(s) to check"
    in_group.add_argument("mol_pred", default=[], type=_path, nargs="*", help=help)
    in_group.add_argument("-l", dest="mol_true", type=_path, help="true molecule, e.g. crystal ligand")
    in_group.add_argument("-p", dest="mol_cond", type=_path, help="conditioning molecule, e.g. protein")
    help = "run multiple inputs listed in a .csv file"
    in_group.add_argument("-t", dest="table", type=_path, help=help)

    # output options
    out_group.add_argument("--outfmt", choices=["short", "long", "csv"], default="short", help="output format")
    out_group.add_argument("--output", type=Path, default=sys.stdout, help="output file (default: stdout)")
    # out_group.add_argument("--snake_case", action="store_false", help="use snake case for output columns")
    out_group.add_argument("--full-report", action="store_true", help="print details for each test")
    out_group.add_argument("--no-header", action="store_true", help="print output without header")

    # config
    cfg_group.add_argument("--config", type=_path, default=None, help="configuration file")
    cfg_group.add_argument(
        "--top-n", type=int, default=None, help="run on TOP_N results in MOL_PRED only (default: all)"
    )
    cfg_group.add_argument(
        "--max-workers",
        type=int,
        default=0,
        help="number workers for parallel processing. (0: single thread, None: all available cores)",
    )
    cfg_group.add_argument(
        "--chunk-size", type=int, help="chunk size for parallel processing of SDF files (default: 100)", default=100
    )

    # other
    inf_group.add_argument("-v", "--version", action="version", version=f"%(prog)s {__version__}")
    inf_group.add_argument("-h", "--help", action="help", help="show this help message and exit")

    namespace = parser.parse_args(args)

    # check that either mol_pred or table was provided
    if namespace.table is None and len(namespace.mol_pred) == 0:
        parser.print_help()
        parser.exit(status=1, message="\nProvide either MOL_PRED or TABLE as input.\n")

    # full report only works with long and csv output
    if namespace.full_report and namespace.outfmt == "short":
        logger.warning("Option --full-report ignored. Please use --outfmt long or csv for --full-report.")
        namespace.full_report = False
    return namespace


def _format_results(df: pd.DataFrame, outfmt: str = "short", no_header: bool = False, index: int = 0) -> str:
    if outfmt == "long":
        return create_long_output(df)

    if outfmt == "csv":
        header = (not no_header) and (index == 0)
        df.index.names = ["file", "molecule", "position"]
        df.columns = [c.lower().replace(" ", "_") for c in df.columns]
        return df.to_csv(index=True, header=header)

    if outfmt == "short":
        return create_short_output(df)

    raise ValueError(f"Unknown output format {outfmt}")


def _select_mode(config: Path | None, columns: Iterable[str]) -> str | dict[str, Any]:
    """Decide which mode to run on based on the inputs."""

    # load config if provided
    if isinstance(config, Path):
        return dict(safe_load(open(config, encoding="utf-8")))

    # forward string if config provide
    if isinstance(config, str):
        return str(config)

    # select mode based on inputs
    if "mol_pred" in columns and "mol_true" in columns and "mol_cond" in columns:
        mode = "redock"
    elif "mol_pred" in columns and "mol_true" in columns:
        mode = "regen"
    elif "mol_pred" in columns and ("protein" in columns) or ("mol_cond" in columns):
        mode = "dock"
    elif any(column in columns for column in ("mol_pred", "mols_pred", "molecule", "molecules", "molecule")):
        mode = "mol"
    else:
        raise NotImplementedError(f"No supported columns found in csv. Columns found are {columns}")

    return mode


def _path(path_str: str | Path) -> Path:
    path = Path(path_str)
    if not path.exists():
        raise argparse.ArgumentTypeError(f"File {path} not found!")
    return path
